from datetime import datetime
import inspect

from django.db import connection
from django.core.cache import cache

from profiler import backends

class ProfilerMiddleware(object):

    def process_request(self, request):
        active_sessions = cache.get('profiler_active_sessions', [])
        #TODO: filter sessions by condition
        backend = backends.make_backend(active_sessions)
        backend.start_request()
        return None

    def process_view(self, request, view_func, view_args, view_kwargs):
        if inspect.ismethod(view_func):
            view_name = view_func.im_class.__module__+ '.' + view_func.im_class.__name__ + view_func.__name__
        else:
            view_name = view_func.__module__ + '.' + view_func.__name__
        backend = backends.get_backend()        
        backend.start_view(view_name)
        return None
    
    def process_response(self, request, response):
        backend = backends.get_backend()
        backend.end_request()
        return response





from datetime import datetime

from django.db.models.sql.compiler import SQLCompiler
from django.db import connection

from profiler.backends import get_backend

def execute_sql(self, *args, **kwargs):
    backend = get_backend()
    if backend is None:
        return self.__execute_sql(*args, **kwargs)
    q, params = self.as_sql()
    start = datetime.now()
    try:
        return self.__execute_sql(*args, **kwargs)
    finally:
        d = (datetime.now() - start)
        backend.log_query(q, d.seconds * 1000 + d.microseconds/1000)
        
INSTRUMENTED = False



if not INSTRUMENTED:
    SQLCompiler.__execute_sql = SQLCompiler.execute_sql
    SQLCompiler.execute_sql = execute_sql
    INSTRUMENTED = True




import threading
import hashlib

import pymongo

from django.core.cache import cache

_hash = lambda x : hashlib.md5(x).hexdigest()


class CacheBackend(object):
    
    def __init__(self, sessions):
        self.sessions = sessions

    def start_request(self):
        pass

    def end_request(self):
        pass
    
    def log_query(self, query, time):
        key = 'profiler_query_%s'%_hash(query)
        query_info = cache.get(key) 
        if query_info is None:
            query_info = { 'times_ran' : 0, 'total_time' : 0, 'query' : query}
            qlist = cache.get('profiler_query_list') or []
            qlist.append(key)
            cache.set('profiler_query_list', qlist)
        query_info['times_ran'] += 1
        query_info['total_time'] += time
        cache.set(key, query_info)
    
    def get_stats(self, session):
        qlist = cache.get('profiler_query_list') or []
        stats = []
        for q in qlist:  
            qdata = cache.get(q) 
            if qdata is None:
                continue
            st = {}
            st.update(qdata)
            st['average_time'] = st['total_time'] / st['times_ran']
            stats += [st]
        return stats
        
connection = pymongo.Connection()
queries = connection['profiler']['queries']

class MongoBackend(object):
    def __init__(self, sessions):
        self.sessions = sessions
        self.view = None

    def start_request(self):
        pass

    def end_request(self):
        pass

    def start_view(self, view):
        self.view = view

    
    def log_query(self, query, time):
        queries.insert({'query' : query,
                        'time' : time,
                        'view' : self.view,
                        'sessions' : self.sessions})
    
    def get_stats(self, session, group_by_view=False):
        if group_by_view:
            group = {'query':2, 'view' : 1}
        else:
            group = {'query':1}
        ret = queries.group(key=group, condition={}, initial={'total_time':0, 'times_ran' : 0}, reduce = 'function(obj, prev){prev.total_time += + obj.time;prev.times_ran +=1;}')
        for r in ret:
            r['average_time'] = r['total_time'] / r['times_ran']
        return ret





_local = threading.local()

def make_backend(sessions):
    _local.backend = MongoBackend(sessions)
    return _local.backend

def get_backend():
    backend = getattr(_local, 'backend', None)
    return backend



from django.http import HttpResponse
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.core.cache import cache
from django.contrib.auth.decorators import user_passes_test

from profiler.backends import get_backend

@user_passes_test(lambda u:u.is_superuser)
def global_stats(request):
    stats = get_backend().get_stats(None)
    return render_to_response('profiler/index.html',
                              {'queries' : stats},
                              context_instance=RequestContext(request))

@user_passes_test(lambda u:u.is_superuser)
def stats_by_view(request):
    stats = get_backend().get_stats(None, group_by_view=True)
    grouped = {}
    for r in stats:
        if r['view'] not in grouped:
            grouped[r['view']] = {'queries' : [], 
                                  'times_ran' : 0,
                                  'total_time' : 0,
                                  'average_time' : 0}
        grouped[r['view']]['queries'].append(r)
        grouped[r['view']]['times_ran'] += r['times_ran']
        grouped[r['view']]['total_time'] += r['total_time']
        grouped[r['view']]['average_time'] += r['average_time']
        
        
    return render_to_response('profiler/by_view.html',
                              {'queries' : grouped},
                              context_instance=RequestContext(request))
